<TeXmacs|1.0.4.6>

<style|mmxdoc>

<\body>
  <tmdoc-title|Symbolic expressions>

  <section|Introduction>

  There are two important ways to implement symbolic expressions for computer
  algebra:

  <\enumerate>
    <item>Implement a ``universal symbolic expression'' type, which is really
    an agglomerate of several concrete types. Every operation on symbolic
    expressions is dispatched to the appropriate concrete type.

    <item>View symbolic expressions types as ``syntactic structures'', in a
    similar way as groups and rings are algebraic structures. Syntactic
    structures should implement constructors for expressions, syntactic
    operations like substitution, as well as an extensible set of other
    operations. Richer syntactic structures can be built on top of simpler
    ones.
  </enumerate>

  The advantage of the first approach, which is used in most classical
  computer algebra systems, is that it is pretty straightforward. In an
  object oriented system, new concrete types can easily be added using
  inheritance, such as in <name|Ginac>.

  However, the first approach one essential drawback: the inheritance system
  is more difficult to use for binary operations when the action to be
  undertaken depends on the types of both arguments. In that case, the
  dispatching has to be implemented by hand and the complexity of the
  approach is bound to increase as more and more concrete types are added.

  The above problem becomes more acute for simultaneous computations with
  non-commutative objects, expressions which auto-expand when multiplying
  them, etc. Although such properties can be elegantly encoded by associating
  a concrete type to each property, one observes a combinatorial explosion of
  the number of cases to consider when multiplying two expressions.

  The main advantage of the second approach is that we regard the whole set
  of operation rules of a symbolic expression type as a syntactic structure
  on itself. This allows to define new symbolic expression types which
  customize simpler ones. For instance, given a symbolic expression type
  <verbatim|Base> in which the multiplication is commutative, we may define a
  new type <verbatim|T> which enriches <verbatim|Base> with expressions which
  do not necessarily commute. Similarly, one may consider a new type with the
  same expressions as <verbatim|Base>, but where some additional
  simplifications are made after certain operations. This is in particular
  useful if we want to quotient <verbatim|Base> by a set of equations or
  rules.

  The main disadvantage of the second approach is that it requires <em|a
  priori> to reimplement all operations of <verbatim|Base> for <verbatim|T>.
  Nevertheless, a great deal of this work can be avoided by an appropriate
  use of template overloading and specialization. However, due to some
  intrinsic shortcomings of<nbsp><verbatim|C++>, it is not clear in all
  circumstances how to proceed.

  <section|The approach in <name|Mmxlib>>

  In <name|Mmxlib>, the main philosophy is to use the second above approach.
  Nevertheless, since we may regard the universal symbolic expression type
  from the first approach as a particular instance of a syntactic structure,
  we may try to incorporate the good sides of the first approach. The main
  question is therefore how to allow for easy customization of symbolic
  expression types.

  First of all, we do use a universal type <verbatim|generic> for the
  <em|representation> of all symbolic expressions. The operations on
  <verbatim|generic> are purely syntactic: no particular rewritings are done
  when adding or multiplying generics. Only a few utilities are included for
  very rudimentary simplifications when printing expressions. Indeed,
  throughout <name|Mmxlib>, the <verbatim|generic> type is also used for
  conversions of types to expressions (and vice versa, in the future).

  At a second stage, the <verbatim|symbolic> template type uses
  <verbatim|generic> as its representation, but reimplements all operations.
  The <verbatim|symbolic> template is parameterized by two types: a type of
  scalars<nbsp><verbatim|C> (which should be regarded as another syntactic
  structure in the future) and a type <verbatim|K> for customization. By
  default, <verbatim|K> is of the form <verbatim|rules\<less\>U\<gtr\>>,
  where <verbatim|U> is a ``universe of rules''. For
  instance,<nbsp><verbatim|U> determines whether the multiplication is
  commutative or distributative and similarly for the multiplication with
  scalars. Notice that it might be useful in the future to rather see
  <verbatim|C> as a parameter of <verbatim|rules> or <verbatim|U> instead of
  <verbatim|symbolic>.

  The default implementations for the methods of <verbatim|symbolic> can be
  redefined for user provided types<nbsp><verbatim|K>. For this to work, the
  type <verbatim|K> should specify static properties
  <verbatim|scalar_commute>, <verbatim|general_commute>,
  <verbatim|scalar_expand>, <verbatim|general_expand>, etc. In that case, the
  usual arithmetic operations are defined by default and they can be
  redefined via template specialization. Moreover, the arithmetic operators
  <verbatim|+>, <verbatim|->, <verbatim|*>, <verbatim|/>, <verbatim|%>,
  <verbatim|pow>, <verbatim|sum> and <verbatim|product> are implemented using
  one level of indirection: these operations respectively call
  <verbatim|add>, <verbatim|sub> or <verbatim|neg>, <verbatim|mul>,
  <verbatim|div>, <verbatim|mod>, <verbatim|power>, <verbatim|big_sum> and
  <verbatim|big_product>. When redefining the arithmetic operations, this
  makes it possible to ``fall back'' to the standard implementation under
  certain circumstances.

  <section|Semantics of symbolic expressions>

  In general, symbolic expressions have a well-defined operational semantics,
  but equality testing is only syntactic. This means that the syntactic
  structure under consideration usually comes with an explicit or implicit
  set of rewriting rules. Although these rules are typically used throughout
  computations so as to simplify expressions, the rules do not necessarily
  lead to normal forms. Consequently, syntactic equality testing is not
  sufficient for testing whether two expressions are equal modulo the current
  rewriting rules. Of course, in certain simple cases (like polynomial
  algebra), an <em|ad hoc>, semantically correct equality test may be
  implemented. Notice also that some of the default rewriting rules may be
  mathematical incorrect in general: <with|mode|math|(x<rsup|y>)<rsup|z>=x<rsup|y*z>>
  may be wrong if <with|mode|math|x=2> and <with|mode|math|y=<frac|1|2>>.

  The current implementation of <verbatim|symbolic\<less\>C,rules\<less\>U\<gtr\>
  \<gtr\>> supports the following rules:

  <\itemize>
    <item>The usual rules for (not necessarily commutative) linear algebra
    over <verbatim|C>.

    <item>Depending on the value of <verbatim|U-\<gtr\>scalar_commute>
    <abbr|resp.> <verbatim|U-\<gtr\>general_commute>, multiplication by
    scalars <abbr|resp.> general multiplication is commutative.

    <item>Depending on the value of <verbatim|U-\<gtr\>scalar_expand>
    <abbr|resp.> <verbatim|U-\<gtr\>general_expand>, multiplication by
    scalars <abbr|resp.> general multiplication is distributive. In case of
    distributivity, we expand <with|mode|math|a*(b+c)=a*b+a*c> and
    <with|mode|math|(a+b)*c=a*c+b*c> when multiplying.

    <item>Depending on the value of <verbatim|U-\<gtr\>scalar_power_add>
    <abbr|resp.> <verbatim|U-\<gtr\>general_power_add>, we perform the
    rewritings <with|mode|math|a<rsup|b>*a<rsup|c>=a<rsup|b+c>> when
    <with|mode|math|b> or <with|mode|math|c> is a scalar, or in general.

    <item>Depending on <verbatim|U-\<gtr\>scalar_power_multiply> <abbr|resp.>
    <verbatim|U-\<gtr\>general_power_multiply>, we perform the rewritings
    <with|mode|math|(a<rsup|b>)<rsup|c>=a<rsup|b*c>> when <with|mode|math|c>
    is a scalar <abbr|resp.> general.

    <item>Depending on <verbatim|U-\<gtr\>scalar_power_expand> <abbr|resp.>
    <verbatim|U-\<gtr\>general_power_expand>, we perform the rewritings
    <with|mode|math|(a*b)<rsup|c>=a<rsup|c>*b<rsup|c>> when
    <with|mode|math|c> is a scalar <abbr|resp.> general.
  </itemize>

  From the implementation point of view, a linear combination
  <with|mode|math|\<lambda\><rsub|1>*a<rsub|1>+\<cdots\>+\<lambda\><rsub|n>*a<rsub|n>>,
  with scalar <with|mode|math|\<lambda\><rsub|i>\<neq\>0> and non-scalar
  <with|mode|math|a<rsub|i>>, is represented by a hash table which associates
  <with|mode|math|\<lambda\><rsub|i>> to each <with|mode|math|a<rsub|i>>. In
  particular, the product <with|mode|math|\<lambda\>*a> of a scalar
  <with|mode|math|\<lambda\>\<nin\>{0,1}> and a non-scalar expression
  <with|mode|math|a> is represented in this way.

  Similarly, a commutative power product <with|mode|math|a<rsub|1><rsup|b<rsub|1>>*\<cdots\>*a<rsub|n><rsup|b<rsub|n>>>,
  with general <with|mode|math|a<rsub|i>\<nin\>{0,1}> and
  <with|mode|math|b<rsub|i>\<neq\>0>, is represented by a hash tables which
  associates <with|mode|math|b<rsub|i>> to each <with|mode|math|a<rsub|i>>.
  Whenever <with|mode|math|a<rsub|i>> is a scalar, then we must have
  <with|mode|math|b<rsub|i>\<nin\>{\<um\>1,<no-break>0,1}>. Non-commutative
  products (if <verbatim|U-<no-break>\<gtr\>general_commute> is false) are
  always of the form <with|mode|math|a<rsub|1>*(a<rsub|2>*\<cdots\>*(a<rsub|n>)
  \<cdots\>)>. Nevertheless, products of a scalar and a non-scalar are still
  represented in the above way.

  <section|To do>

  At the moment, we really took care of rational scalars. Special attention
  should be paid to scalars which are syntactic structures. We might define a
  <verbatim|property\<less\>T\<gtr\>> template type which may be used to
  associate properties to types <verbatim|T>, like
  <verbatim|syntactic_structure>, <verbatim|ring>, etc.

  Complex computations, like repeated differentiation, often lead to
  expressions with a large ``polynomial part''. It would be good to detect
  automatically when the polynomial expansion of a subexpression in a
  variable <with|mode|math|x> leads to simplifications. This could for
  instance be done by keeping track of the minimal and maximal degrees of all
  variables <with|mode|math|x> occurring in the subexpression. Notice that
  this may require rewriting expressions as ``polynomial expressions'', which
  requires in particular rewritings of the form
  <with|mode|math|\<mathe\><rsup|2*x+y>\<rightarrow\>(\<mathe\><rsup|x>)<rsup|2>*\<mathe\><rsup|y>>.
  Notice that it may be simpler to directly perform simplifications of this
  kind in a more <em|ad hoc> way during operations like differentiation (by
  carefully treating the differentiation of power products).

  <tmdoc-copyright|2005|Joris van der Hoeven>

  <tmdoc-license|Permission is granted to copy, distribute and/or modify this
  document under the terms of the <hlink|GNU General Public
  License|http://www.gnu.org/licenses/gpl.txt>. If you don't have this file,
  write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.>
</body>