
/******************************************************************************
* MODULE     : analytic.hpp
* DESCRIPTION: Analytic functions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_ANALYTIC_HPP
#define __MML_ANALYTIC_HPP
#include "complex.hpp"
#include "numeric_polynomial.hpp"
#include "linear_series.hpp"
namespace mml {
#define TMPL_DEF template<typename V>
#define TMPL template<typename V>
#define Analytic analytic<V>
#define Analytic_rep analytic_rep<V>
#define Positive typename V::positive_type
#define Complex typename V::complex_type
#define Series typename V::series_type
TMPL class analytic;

/******************************************************************************
* Different variants of analytic functions
******************************************************************************/

class double_analytic {
public:
  typedef double positive_type;
  typedef complexify<double> complex_type;
  typedef series<complex_type> series_type;
  static inline positive_type _plus_infinity () { return HUGE_VAL; }
  static inline positive_type _abs (const complex_type& z) {
    return mml::abs (z); }
  static inline complex_type _construct (const positive_type& x) { return x; }
};

inline bool is_infinity (double x) { return x >= HUGE_VAL; }

class exact_analytic {
public:
  typedef real_ball positive_type;
  typedef complex complex_type;
  typedef series<complex_type> series_type;
  static inline positive_type _plus_infinity () {
    return Plus_infinity (positive_type); }
  static positive_type _abs (const complex_type& z) {
    return upper (current (hypot (Re (z), Im (z)))); } // CHECK
  static inline complex_type _construct (const positive_type& x) {
    return gaussian (real (center (x)), real (0)); } // CHECK
};

/******************************************************************************
* The analytic function class
******************************************************************************/

TMPL_DEF
class analytic_rep REP_STRUCT {
public:
  nat complexity;

protected:
  //inline analytic_rep () {}
  inline analytic_rep (): complexity (0) {}
  inline virtual ~analytic_rep () {}

public:
  virtual generic expression (const generic& x) const = 0;
  virtual Series germ () const = 0;
  virtual Positive radius (nat prec= 1) const = 0;
  virtual Positive bound (const Positive& r, nat prec= 1) const = 0;
  virtual Analytic move (const Complex& z) const = 0;
  virtual Complex eval (const Complex& z) const;
  Positive root_free_disk (nat prec= 1) const;
  Positive lower_bound (const Positive& r, nat prec= 1) const;
  Analytic continuate (const Complex& z) const;
  Complex evaluate (const Complex& z) const;
  friend class Analytic;
};

TMPL_DEF
class analytic {
INDIRECT_PROTO_1 (analytic, analytic_rep, V)
public:
  analytic (int c= 0);
  analytic (const Complex& c);
  analytic (const polynomial<Complex>& P);
  inline Complex operator [] (nat i) const { return rep->germ () [i]; }
  inline Complex operator () (const Complex& z) const { return rep->eval (z); }
};
INDIRECT_IMPL_1 (analytic, analytic_rep, typename V, V)

TMPL generic var (const Analytic& f) { return var (f->germ ()); }

TMPL Analytic operator + (const Analytic& f, const Analytic& g);
TMPL Analytic operator - (const Analytic& f);
TMPL Analytic operator - (const Analytic& f, const Analytic& g);
TMPL Analytic operator * (const Complex& c, const Analytic& f);
TMPL Analytic operator * (const Analytic& f, const Complex& c);
TMPL Analytic operator / (const Analytic& f, const Complex& c);
TMPL Analytic operator * (const Analytic& f, const Analytic& g);
TMPL Analytic operator / (const Analytic& f, const Analytic& g);

TMPL Analytic derive (const Analytic& f);
TMPL Analytic xderive (const Analytic& f);
TMPL Analytic integrate (const Analytic& f, const Complex& c= Complex (0));
TMPL Analytic exp (const Analytic& f);
TMPL Analytic log (const Analytic& f);
TMPL Analytic log_init (const Analytic& f, const Complex& c);
TMPL Analytic solve_lde (const vector<Analytic >& f,
			 const vector<Complex >& c);

/******************************************************************************
* Helper functions
******************************************************************************/

template<typename Real> Real
radius_between (const Real& r, const Real& R) {
  if (is_infinity (R)) return Real (2) * r;
  else return (r + R) / Real (2);
}

/******************************************************************************
* Default routines for analytic functions
******************************************************************************/

TMPL Complex
Analytic_rep::eval (const Complex& z) const {
  return move (z) [0];
}

TMPL Positive
Analytic_rep::root_free_disk (nat prec) const {
  Positive C = V::_abs (germ () [0]);
  Positive R0= min (radius (prec), Positive (1 << prec));
  Positive R = R0 / 4;
  Positive S = R0 / 2;
  Positive D = bound (S) * S / square (S - R);
  return min (R, C / D);
}

TMPL Positive
Analytic_rep::lower_bound (const Positive& R, nat prec) const {
  Positive C = V::_abs (germ () [0]);
  Positive R0= min (radius (prec), Positive (1 << prec));
  Positive S = (2 * R + R0) / 3;
  Positive D = bound (S) * S / square (S - R);
  return C - R * D;
}

template<> exact_analytic::positive_type
analytic_rep<exact_analytic>::root_free_disk (nat prec) const;

template<> exact_analytic::positive_type
analytic_rep<exact_analytic>::lower_bound 
  (const exact_analytic::positive_type& r, nat prec) const;

TMPL Analytic
Analytic_rep::continuate (const Complex& z) const {
  Positive R= radius ();
  Positive Z= V::_abs (z);
  if (Positive(3)*Z < R) {
    // NOTE: instead of Z < R, so as to avoid slow convergence
    Analytic ff= move (z);
    ff->complexity= complexity + 1;
    return ff;
    //return move (z);
  }
  else {
    Complex u= V::_construct (R / (Positive (8) * Z)) * z;
    // Complex u= V::_construct (R / (Positive (3) * Z)) * z;
    // NOTE: instead of 2*Z, for faster convergence
    // cout << "{" << z-u << "}\n"; cout.flush ();
    Analytic ff= move (u);
    ff->complexity= complexity + 1;
    return ff -> continuate (z-u);
    //return move (u) -> continuate (z-u);
  }
}

/******************************************************************************
* Evaluation of analytic functions
******************************************************************************/

TMPL Complex
Analytic_rep::evaluate (const Complex& z) const {
  assert (false); // only implemented for specific variants
}

template<> double_analytic::complex_type
analytic_rep<double_analytic>::evaluate
  (const double_analytic::complex_type& z) const;

template<> exact_analytic::complex_type
analytic_rep<exact_analytic>::evaluate
  (const exact_analytic::complex_type& z) const;

/******************************************************************************
* Hashing and printing analytic functions
******************************************************************************/

TMPL nat
hash (const Analytic& f) {
  Series s= f->germ ();
  register nat i, h= 7531;
  for (i=0; i<4; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash (Complex (s[i]));
  return h;
}

TMPL generic
as_generic (const Analytic& f) {
  Series s= f->germ ();
  if (get_series_formula_output ()) return f->expression (var (f));
  generic sum= 0;
  nat order= get_series_precision ();
  for (nat i=0; i<order; i++)
    sum= plus (sum, times (as_generic (s[i]), power (var (f), generic (i))));
  sum= plus (sum, generic ("O") (power (var (f), generic (order))));
  return sum;
}

PRINT_SUGAR(TMPL,Analytic)

/******************************************************************************
* Explicit analytic functions
******************************************************************************/

TMPL
class polynomial_analytic_rep: public Analytic_rep {
  polynomial<Complex> P;
public:
  polynomial_analytic_rep (const polynomial<Complex>& P2): P (P2) {}
  generic expression (const generic& x) const {
    (void) x;
    return as_generic (P); }
  Series germ () const {
    return P; }
  Positive radius (nat prec) const {
    (void) prec;
    return V::_plus_infinity (); }
  Positive bound (const Positive& r, nat prec) const {
    (void) prec;
    Positive sum (0), prod (1);
    for (nat i=0; i<dim(P); i++) {
      sum += V::_abs (P[i]) * prod;
      prod *= r;
    }
    return sum;
  }
  Analytic move (const Complex& z) const {
    polynomial<Complex> Q= seq (z, Complex (1));
    return Analytic (compose (P, Q));
  }
};

TMPL
Analytic::analytic (int c) {
  rep= new polynomial_analytic_rep<V> (polynomial<Complex> (Complex (c), 0));
}

TMPL
Analytic::analytic (const Complex& c) {
  rep= new polynomial_analytic_rep<V> (polynomial<Complex> (c, 0));
}

TMPL
Analytic::analytic (const polynomial<Complex>& P) {
  rep= new polynomial_analytic_rep<V> (P);
}

/******************************************************************************
* Sums and differences
******************************************************************************/

TMPL
class add_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  add_analytic_rep (const Analytic& f2, const Analytic& g2):
    f(f2), g(g2) {}
  generic expression (const generic& x) const {
    return plus (f->expression (x), g->expression (x)); }
  Series germ () const {
    return f->germ () + g->germ (); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->radius (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) + g->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) + g->move (z); }
};

TMPL Analytic
operator + (const Analytic& f, const Analytic& g) {
  return Analytic (new add_analytic_rep<V> (f, g));
}

TMPL
class sub_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  sub_analytic_rep (const Analytic& f2, const Analytic& g2):
    f(f2), g(g2) {}
  generic expression (const generic& x) const {
    return minus (f->expression (x), g->expression (x)); }
  Series germ () const {
    return f->germ () - g->germ (); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->radius (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) + g->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) - g->move (z); }
};

TMPL Analytic
operator - (const Analytic& f) {
  return Analytic (new sub_analytic_rep<V> (0, f));
}

TMPL Analytic
operator - (const Analytic& f, const Analytic& g) {
  return Analytic (new sub_analytic_rep<V> (f, g));
}

/******************************************************************************
* Scalar multiplication and division
******************************************************************************/

TMPL
class scmul_analytic_rep: public Analytic_rep {
  Complex  c;
  Analytic f;
public:
  scmul_analytic_rep (const Complex& c2, const Analytic& f2):
    c (c2), f (f2) {}
  generic expression (const generic& x) const {
    return times (as_generic (c), f->expression (x)); }
  Series germ () const {
    return c * f->germ (); }
  Positive radius (nat prec) const {
    return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    return V::_abs (c) * f->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return c * f->move (z); }
};

TMPL Analytic
operator * (const Complex& c, const Analytic& f) {
  return Analytic (new scmul_analytic_rep<V> (c, f));
}

TMPL Analytic
operator * (const Analytic& f, const Complex& c) {
  return Analytic (new scmul_analytic_rep<V> (c, f));
}

TMPL Analytic
operator / (const Analytic& f, const Complex& c) {
  return Analytic (new scmul_analytic_rep<V> (1/c, f));
}

/******************************************************************************
* Multiplication
******************************************************************************/

TMPL
class mul_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  mul_analytic_rep (const Analytic& f2, const Analytic& g2):
    f (f2), g (g2) {}
  generic expression (const generic& x) const {
    return times (f->expression (x), g->expression (x)); }
  Series germ () const {
    return f->germ () * g->germ (); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->radius (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) * g->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) * g->move (z); }
};

TMPL Analytic
operator * (const Analytic& f, const Analytic& g) {
  return Analytic (new mul_analytic_rep<V> (f, g));
}

/******************************************************************************
* Division
******************************************************************************/

TMPL
class div_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  div_analytic_rep (const Analytic& f2, const Analytic& g2):
    f (f2), g (g2) {}
  generic expression (const generic& x) const {
    return over (f->expression (x), g->expression (x)); }
  Series germ () const {
    return f->germ () / g->germ (); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->root_free_disk (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) / g->lower_bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) / g->move (z); }
};

TMPL Analytic
operator / (const Analytic& f, const Analytic& g) {
  // FIXME: add similar valuation-based stuff as for Series
  return Analytic (new div_analytic_rep<V> (f, g));
}

/******************************************************************************
* Differentiation
******************************************************************************/

TMPL
class derive_analytic_rep: public Analytic_rep {
  Analytic f;
public:
  derive_analytic_rep (const Analytic& f2): f (f2) {}
  generic expression (const generic& x) const {
    return as_generic ("derive", f->expression (var (f)), var (f)) (x); }
  Series germ () const { return derive (f->germ ()); }
  Positive radius (nat prec) const { return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    Positive R= radius_between (r, radius (prec));
    return (R / square (R - r)) * f->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return derive (f->move (z)); }
};

TMPL Analytic
derive (const Analytic& f) {
  return Analytic (new derive_analytic_rep<V> (f));
}

TMPL Analytic
xderive (const Analytic& f) {
  return Analytic (polynomial<Complex > (Complex (1), 1)) * derive (f);
}

/******************************************************************************
* Integration
******************************************************************************/

TMPL
class integrate_analytic_rep: public Analytic_rep {
  Analytic f;
  Complex  c;
public:
  integrate_analytic_rep (const Analytic& f2, const Complex& c2):
    f (f2), c (c2) {}
  generic expression (const generic& x) const {
    return as_generic ("integrate", f->expression (var (f)), var (f)) (x); }
  Series germ () const { return integrate (f->germ ()) + Series (c); }
  Positive radius (nat prec) const { return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    return r * f->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return integrate (f->move (z), evaluate (z)); }
  Complex eval (const Complex& z) const {
    return evaluate (z); }
};

TMPL Analytic
integrate (const Analytic& f, const Complex& c) {
  return Analytic (new integrate_analytic_rep<V> (f, c));
}

/******************************************************************************
* Exponentiation
******************************************************************************/

TMPL
class exp_analytic_rep: public Analytic_rep {
  Analytic f;
public:
  exp_analytic_rep (const Analytic& f2): f (f2) {}
  generic expression (const generic& x) const {
    return as_generic ("exp", f->expression (x)); }
  Series germ () const { return exp (f->germ ()); }
  Positive radius (nat prec) const { return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    return exp (f->bound (r, prec)); }
  Analytic move (const Complex& z) const {
    return exp (f->move (z)); }
};

TMPL Analytic
exp (const Analytic& f) {
  return Analytic (new exp_analytic_rep<V> (f));
}

/******************************************************************************
* Logarithm
******************************************************************************/

TMPL
class log_analytic_rep: public Analytic_rep {
  Analytic f;
  Complex  c;
public:
  log_analytic_rep (const Analytic& f2, const Complex& c2):
    f (f2), c (c2) {}
  generic expression (const generic& x) const {
    return as_generic ("log", f->expression (x), c); }
  Series germ () const { return log_init (f->germ (), c); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), f->root_free_disk (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return max (log (f->bound (r, prec)), log (f->lower_bound (r, prec))); }
  Analytic move (const Complex& z) const {
    Analytic ff= f->move (z);
    return log_init (ff, c + log (ff[0] / f[0])); }
};

TMPL Analytic
log (const Analytic& f) {
  return Analytic (new log_analytic_rep<V> (f, log (f[0])));
}

TMPL Analytic
log_init (const Analytic& f, const Complex& c) {
  return Analytic (new log_analytic_rep<V> (f, c));
}

/******************************************************************************
* Linear differential equations
******************************************************************************/

TMPL vector<Series >
as_vector_series (const vector<Analytic >& f) {
  nat i, n= dim (f);
  vector<Series > v (Series (0), n);
  for (i=0; i<n; i++)
    v(i)= f[i]->germ ();
  return v;
}

TMPL
class solve_lde_analytic_rep: public Analytic_rep {
  vector<Analytic > f;
  vector<Complex  > c;
public:
  solve_lde_analytic_rep (const vector<Analytic >& f2,
			  const vector<Complex  >& c2):
    f (f2), c (c2) {}
  generic expression (const generic& x) const {
    return as_generic ("solve_lde", f, c); }
  Series germ () const {
    return solve_lde (as_vector_series (f), c); }
  Positive radius (nat prec) const {
    Positive r= V::_plus_infinity ();
    for (nat i= 0; i < dim (f); i++)
      r= min (f[i]->radius (prec), r);
    return r; }
  Positive bound (const Positive& R, nat prec) const {
    Positive R0= min (radius (prec), R * (1 << (prec+1)));
    Positive S = (R + R0) / 2;
    Positive C = 0;
    Positive M = 0;
    for (nat i= 0; i < dim (f); i++) {
      C= max (C, V::_abs (c[i]));
      M= max (M, f[i]->bound (S));
    }
    return C * exp (log (S / (S - R)) * (M+1) * S); }
  Analytic move (const Complex& z) const {
    Analytic dfs (this, true);
    vector<Analytic > ff (Analytic (0), dim (f));
    vector<Complex  > cc (Complex  (0), dim (f));
    for (nat i= 0; i < dim (f); i++) {
      ff(i)= f[i]->move (z);
      cc(i)= dfs->evaluate (z);
      dfs  = derive (dfs);
      dfs->complexity= this->complexity;
    }
    return solve_lde (ff, cc); }
  Complex eval (const Complex& z) const {
    return evaluate (z); }
};

TMPL Analytic
solve_lde (const vector<Analytic >& f, const vector<Complex >& c) {
  /*
  cout << set_series_expr (true)
       << "[Solve: " << f << ", " << c << "]"
       << set_series_expr (false); cout.flush ();
  */
  return Analytic (new solve_lde_analytic_rep<V> (f, c));
}

#undef TMPL_DEF
#undef TMPL
#undef Analytic
#undef Analytic_rep
#undef Positive
#undef Complex
#undef Series
} // namespace mml
#endif // __MML_ANALYTIC_HPP
