
/******************************************************************************
* MODULE     : certify.hpp
* DESCRIPTION: Certified numeric computations
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_CERTIFY_HPP
#define __MML_CERTIFY_HPP
#include "rational.hpp"
#include "type_props.hpp"
#include "generic_extra.hpp"
#include "operators.hpp"
namespace mml {
#define TMPL_DEF template<typename C,typename V=std_certify>
#define TMPL template<typename C,typename V>
#define Certify certify<C,V>
#define Certify_rep certify_rep<C,V>
#define Center Center_type(C)
#define Radius Radius_type(C)
#define Real_certify certify<Real_type(C),V>
#define Real_certify_rep certify_rep<Real_type(C),V>
#define Abs_certify certify<Abs_type(C),V>
#define Abs_certify_rep certify_rep<Abs_type(C),V>
#define Complex_certify certify<Complex_type(C),V>
#define Complex_certify_rep certify_rep<Complex_type(C),V>
TMPL class certify;
TMPL inline C current (const Certify& z);
TMPL C approximate (const Certify& z, const Radius& err, bool tf= true);
TMPL C normalize (const Certify& z, const Radius& rel_err, bool tf= true);

/******************************************************************************
* Global variables
******************************************************************************/

extern nat  mml_cancel_order;    // bit precision for zero testing
extern bool mml_formula_output;  // print formulas vs. approximations

inline nat set_cancel_order (nat prec) {
  nat old= mml_cancel_order; mml_cancel_order= prec; return old; }
inline void reset_cancel_order (nat prec) { mml_cancel_order= prec; }
inline nat get_cancel_order () { return mml_cancel_order; }

inline bool set_formula_output (bool mode) {
  bool old= mml_formula_output; mml_formula_output= mode; return old; }
inline void reset_formula_output (bool mode) { mml_formula_output= mode; }
inline bool get_formula_output () { return mml_formula_output; }

/******************************************************************************
* Variants for the certify class
******************************************************************************/

class std_certify {};

/******************************************************************************
* The certify class
******************************************************************************/

TMPL_DEF
class certify_rep REP_STRUCT {
protected:
  nat prec;   // precision with which best was computed
  C   best;   // best currently available approximation

public:
  inline certify_rep (): prec (DEFAULT_LOW), best (Nan (C)) {}
  virtual inline ~certify_rep () {}

  virtual generic expression () const = 0;
  virtual C compute () = 0;
  virtual inline generic raw_expression () const {
    bool old= set_formula_output (true);
    generic r= expression ();
    reset_formula_output (old);
    return r; }

  inline void init () {
    nat old= set_precision (prec);
    best= compute ();
    reset_precision (old); }
  C compute (nat new_prec) const {
    if (new_prec <= prec) return best;
    if ((prec<<1) > new_prec) {
      nat p= DEFAULT_LOW;
      while (p <= new_prec) p <<= 1;
      new_prec= p;
    }
    const_cast<Certify_rep*> (this)->prec= new_prec;
    nat old= set_precision (prec);
    const_cast<Certify_rep*> (this)->best=
      const_cast<Certify_rep*> (this)->compute ();
    reset_precision (old);
    return best; }

  friend class Certify;
  friend C current LESSGTR (const Certify& z);
  friend C approximate LESSGTR (const Certify& z, const Radius& err, bool tf);
  friend C normalize LESSGTR (const Certify& z, const Radius& rerr, bool tf);
};

TMPL_DEF
class certify {
INDIRECT_PROTO_2 (certify, certify_rep, C, V)
public:
  certify ();
  template<typename T> certify (const T& z);
  template<typename T> certify (const certify<T,V>& z);
  template<typename XT, typename YT> certify (const XT& x, const YT& y);

  friend C current LESSGTR (const Certify& z);
  friend C approximate LESSGTR (const Certify& z, const Radius& err, bool tf);
  friend C normalize LESSGTR (const Certify& z, const Radius& rerr, bool tf);
};
INDIRECT_IMPL_2 (certify, certify_rep, typename C, C, typename V, V)

STYPE_TO_TYPE(TMPL,center_type,Certify,Center);
STYPE_TO_TYPE(TMPL,radius_type,Certify,Radius);
STYPE_TO_TYPE(TMPL,real_type,Certify,Real_certify);
STYPE_TO_TYPE(TMPL,abs_type,Certify,Abs_certify);
STYPE_TO_TYPE(TMPL,complex_type,Certify,Complex_certify);

/******************************************************************************
* Approximation routines
******************************************************************************/

TMPL inline C
current (const Certify& z) {
  return z->best;
}

TMPL C
approximate (const Certify& z, const Radius& err, bool tf) {
  while (!is_number (radius (z->best))) {
    if (z->prec >= get_cancel_order ()) return z->best;
    (void) z.rep->compute (min (z->prec << 1, get_cancel_order ()));
  }
  while (radius (z->best) >= err) {
    nat new_prec= z->prec + exponent (Radius(2.001) * radius(z->best) / err);
    if (new_prec > (z->prec << 4))
      new_prec= min (z->prec << 8, new_prec) >> 4;
    new_prec= ((new_prec + BITS_PER_LIMB-1) / BITS_PER_LIMB) * BITS_PER_LIMB;
    (void) z.rep->compute (new_prec);
    if (radius (z->best) >= err) {
      cout << "[xfactor=" << radius (z->best) / err << "]";
      cout.flush ();
    }
    //assert (radius (z->best) < err); // when using Lipschitz balls
  }
  if (tf && is_number (z->best)) {
    nat old= set_precision (z->prec);
    C ret (center (z->best), err);
    reset_precision (old);
    return ret;
  }
  else return z->best;
}

TMPL C
normalize (const Certify& z, const Radius& rel_err, bool tf) {
  while (radius (z->best) >= mul_D (Radius (0.0001), abs_D (z->best))) {
    if (z->prec >= get_cancel_order ()) {
      if (is_number (z->best)) return 0;
      else return z->best;
    }
    (void) z.rep->compute (min (z->prec << 1, get_cancel_order ()));
  }
  return approximate (z, rel_err * abs_U (z->best), tf);
}

TMPL nat
hash (const Certify& z) {
  C r= normalize (z, Radius (1) >> (BITS_PER_LIMB+1), false);
  return hash (r);
}

TMPL C
expand (const Certify& z, nat prec= get_cancel_order (), bool tf= true) {
  return normalize (z, Radius (1) >> prec, tf);
}

TMPL inline bool is_number (const Certify& z) {
  return is_nan (normalize (z, Radius (0.25))); }
TMPL inline bool is_infinity (const Certify& z) {
  return is_infinity (normalize (z, Radius (0.25))); }
TMPL inline bool is_fuzz (const Certify& z) {
  return is_fuzz (normalize (z, Radius (0.25))); }
TMPL inline bool is_nan (const Certify& z) {
  return is_nan (normalize (z, Radius (0.25))); }
TMPL inline bool is_zero (const Certify& z) {
  return normalize (z, Radius (0.25)) == 0; }
TMPL inline bool is_non_zero (const Certify& z) {
  return normalize (z, Radius (0.25)) != 0; }
TMPL inline int sign (const Certify& z) {
  return sign (normalize (z, Radius (0.25))); }
TMPL inline int var_sign (const Certify& z) {
  return var_sign (normalize (z, Radius (0.25))); }

TMPL inline bool operator == (const Certify& x, const Certify& y) {
  return is_zero (x-y); }
TMPL inline bool operator != (const Certify& x, const Certify& y) {
  return is_non_zero (x-y); }
TMPL inline bool operator <= (const Certify& x, const Certify& y) {
  return sign (x-y) <= 0; }
TMPL inline bool operator <  (const Certify& x, const Certify& y) {
  return sign (x-y) <  0; }
TMPL inline bool operator >= (const Certify& x, const Certify& y) {
  return sign (x-y) >= 0; }
TMPL inline bool operator >  (const Certify& x, const Certify& y) {
  return sign (x-y) >  0; }

TMPL generic
as_generic (const Certify& z) {
  if (get_formula_output ()) return z->expression ();
  else if (is_nan (z)) return "NaN";
  else if (is_infinity (z)) return as_generic (current (z));
  else if (is_zero (z)) return 0;
  else return as_generic (expand (z, 20), 4);
}

PRINT_SUGAR(TMPL,Certify);
COMPARE_INT_SUGAR(TMPL,Certify,Certify);

/******************************************************************************
* Constructors
******************************************************************************/

template<typename C, typename V, typename T>
class constant_certify_rep: public Certify_rep {
  T c;
public:
  inline constant_certify_rep (const T& c2): c (c2) { this->init (); }
  generic expression () const { return as_generic (c); }
  C compute () { return C (c); }
};

TMPL inline
Certify::certify (): rep (new constant_certify_rep<C,V,int> (0)) {}

TMPL template<typename T> inline
Certify::certify (const T& c): rep (new constant_certify_rep<C,V,T> (c)) {}

template<typename C, typename V, typename T>
class cast_certify_rep: public Certify_rep {
  certify<T,V> z;
public:
  inline cast_certify_rep (const certify<T,V>& z2):
    z (z2) { this->init (); }
  generic expression () const { return as_generic (z); }
  C compute () { return C (z->compute (this->prec)); }
};

TMPL template<typename T> inline
Certify::certify (const certify<T,V>& z):
  rep (new cast_certify_rep<C,V,T> (z)) {}

/******************************************************************************
* Constants
******************************************************************************/

template<typename Op,typename C,typename V>
class cst_certify_rep: public Certify_rep {
public:
  inline cst_certify_rep () { this->init (); }
  generic expression () const { return Op::template op<generic> (); }
  C compute () { return Op::template op<C> (); }
};

template<typename Op,typename C,typename V>
inline Certify
cst_certify () {
  return (Certify_rep*) new cst_certify_rep<Op,C,V> (); }

TMPL
struct constant_helper<Certify > {
  static inline Certify log2 () {
    return cst_certify<log2_op,C,V> (); }
  static inline Certify pi () {
    return cst_certify<pi_op,C,V> (); }
  static inline Certify euler () {
    return cst_certify<euler_op,C,V> (); }
  static inline Certify imaginary () {
    return cst_certify<imaginary_op,C,V> (); }
  static inline Certify infinity () {
    return cst_certify<infinity_op,C,V> (); }
  static inline Certify plus_infinity () {
    return cst_certify<plus_infinity_op,C,V> (); }
  static inline Certify minus_infinity () {
    return cst_certify<minus_infinity_op,C,V> (); }
};

/******************************************************************************
* Abstract operations
******************************************************************************/

template<typename Op,typename C,typename V>
class unary_certify_rep: public Certify_rep {
  Certify x;
public:
  inline unary_certify_rep (const Certify& x2): x (x2) { this->init (); }
  generic expression () const { return Op::op (as_generic (x)); }
  C compute () { return Op::op (x->compute (this->prec)); }
};

template<typename Op,typename C,typename V> inline Certify
unary_certify (const Certify& x) {
  return (Certify_rep*) new unary_certify_rep<Op,C,V> (x); }

template<typename Op,typename C,typename V>
class binary_certify_rep: public Certify_rep {
  Certify x, y;
public:
  inline binary_certify_rep (const Certify& x2, const Certify& y2):
    x (x2), y (y2) { this->init (); }
  generic expression () const {
    return Op::op (as_generic (x), as_generic (y)); }
  C compute () {
    return Op::op (x->compute (this->prec), y->compute (this->prec)); }
};

template<typename Op,typename C,typename V> inline Certify
binary_certify (const Certify& x, const Certify& y) {
  return (Certify_rep*) new binary_certify_rep<Op,C,V> (x, y); }

template<typename Op,typename C,typename V,typename X>
class extrary_certify_rep: public Certify_rep {
  Certify x;
  X p;
public:
  inline extrary_certify_rep (const Certify& x2, const X& p2):
    x (x2), p (p2) { this->init (); }
  generic expression () const {
    return Op::op (as_generic (x), as_generic (p)); }
  C compute () {
    return Op::op (x->compute (this->prec), p); }
};

template<typename Op,typename C,typename V,typename X> inline Certify
extrary_certify (const Certify& x, const X& p) {
  return (Certify_rep*) new extrary_certify_rep<Op,C,V,X> (x, p); }

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL Certify inline operator - (const Certify& x) {
  return unary_certify<neg_op> (x); }
TMPL Certify inline operator + (const Certify& x, const Certify& y) {
  return binary_certify<add_op> (x, y); }
TMPL Certify inline operator - (const Certify& x, const Certify& y) {
  return binary_certify<sub_op> (x, y); }
TMPL Certify inline operator * (const Certify& x, const Certify& y) {
  return binary_certify<mul_op> (x, y); }
TMPL Certify inline operator / (const Certify& x, const Certify& y) {
  return binary_certify<div_op> (x, y); }
TMPL Certify inline operator << (const Certify& x, int sh) {
  return extrary_certify<lsh_op<false> > (x, sh); }
TMPL Certify inline operator >> (const Certify& x, int sh) {
  return extrary_certify<rsh_op<false> > (x, sh); }

ARITH_INT_SUGAR(TMPL,Certify,Certify);
SHIFT_INT_SUGAR(TMPL,Certify,Certify);

/******************************************************************************
* Special functions
******************************************************************************/

TMPL Certify exp (const Certify& x) { return unary_certify<exp_op> (x); }
TMPL Certify log (const Certify& x) { return unary_certify<log_op> (x); }
TMPL Certify cos (const Certify& x) { return unary_certify<cos_op> (x); }
TMPL Certify sin (const Certify& x) { return unary_certify<sin_op> (x); }
TMPL Certify tan (const Certify& x) { return unary_certify<tan_op> (x); }
TMPL Certify atan (const Certify& x) { return unary_certify<atan_op> (x); }

TMPL Certify
sqrt (const Certify& x) {
  if (x == 0) return 0;
  return unary_certify<sqrt_op> (x);
}

TMPL Certify
acos (const Certify& x) {
  if (x == -1) return Pi(Certify);
  if (x ==  1) return 0;
  return unary_certify<acos_op> (x);
}

TMPL Certify
asin (const Certify& x) {
  if (x == -1) return -Pi(Certify) >> 1;
  if (x ==  1) return Pi(Certify) >> 1;
  return unary_certify<asin_op> (x);
}

TMPL Certify
pow (const Certify& x, const Certify& y) {
  return exp (log (x) * y);
}

TMPL Certify
atan2 (const Certify& y, const Certify& x) {
  assert (max (abs (x), abs (y)) > 0);
  if (sign (lower (current (x))) > 0) return atan (y/x);
  if (sign (lower (current (y))) > 0) return atan (-x/y) + (Pi(Certify) >> 1);
  if (sign (upper (current (y))) < 0) return atan (-x/y) - (Pi(Certify) >> 1);
  if (sign (upper (current (x))) < 0) return atan (y/x) + Pi(Certify);
  assert (false); // this situation should never occur
  return 0;
}

/******************************************************************************
* Further routines involving comparisons
******************************************************************************/

TMPL
class abs_rep: public Abs_certify_rep {
  Certify  z;
public:
  inline abs_rep (const Certify& z2): z (z2) { this->init (); }
  generic expression () const { return abs (as_generic (z)); }
  Abs_type(C) compute () { return abs (z->compute (this->prec)); }
};

TMPL inline Abs_certify abs (const Certify& z) {
  return (Abs_certify_rep*) new abs_rep<C,V> (z); }

TMPL integer
round (const Certify& z, int direction) {
  integer i;
  Center r= center (approximate (z, 1e-20));
  switch (direction) {
  case -1: i= as_integer (floor (r)); break;
  case  0: i= as_integer (trunc (r)); break;
  case  1: i= as_integer (ceil  (r)); break;
  }
  int sgn= sign (z - Certify (i));
  switch (direction) {
  case -1: return sgn<0? i - integer(1): i;
  case  0:
    if (sgn<0 && i>0) return i - integer (1);
    else if (sgn>0 && i<0) return i + integer (1);
    else return i;
  case  1: return sgn>0? i + integer (1): i;
  }
  return i;
}

TMPL inline integer floor (const Certify& x) { return round (x, -1); }
TMPL inline integer trunc (const Certify& x) { return round (x, 0); }
TMPL inline integer ceil  (const Certify& x) { return round (x, 1); }
TMPL integer round (const Certify& x) { return floor (x + (Certify(1)>>1)); }

#undef TMPL_DEF
#undef TMPL
#undef Certify
#undef Certify_rep
#undef Center
#undef Radius
#undef Real_certify
#undef Real_certify_rep
#undef Abs_certify
#undef Abs_certify_rep
#undef Complex_certify
#undef Complex_certify_rep
}
#endif // __MML_CERTIFY_HPP
