
/******************************************************************************
* MODULE     : complexify_certify.hpp
* DESCRIPTION: Certified numeric computations with complexified numbers
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_COMPLEXIFY_CERTIFY_HPP
#define __MML_COMPLEXIFY_CERTIFY_HPP
#include "complexify.hpp"
#include "certify.hpp"
namespace mml {
#define TMPL_DEF template<typename C,typename V=std_certify>
#define TMPL template<typename C,typename V>
#define Certify certify<C,V>
#define Certify_rep certify_rep<C,V>
#define Center Center_type(C)
#define Radius Radius_type(C)
#define Real_certify certify<Real_type(C),V>
#define Real_certify_rep certify_rep<Real_type(C),V>
#define Abs_certify certify<Abs_type(C),V>
#define Abs_certify_rep certify_rep<Abs_type(C),V>
#define Complex_certify certify<Complex_type(C),V>
#define Complex_certify_rep certify_rep<Complex_type(C),V>

/******************************************************************************
* Standard functions
******************************************************************************/

TMPL nat
hash (const Complex_certify& z) {
  C r= normalize (abs (z), Radius (1) >> (BITS_PER_LIMB+1), false);
  return hash (r);
}

/******************************************************************************
* From real numbers to complex numbers
******************************************************************************/

TMPL_DEF
class complex_certify_rep: public Certify_rep {
  Real_certify x1, x2;
public:
  inline complex_certify_rep (const Real_certify& X1, const Real_certify& X2):
    x1 (X1), x2 (X2) { this->init (); }
  generic expression () const {
    return gaussian (as_generic (x1), as_generic (x2)); }
  C compute () {
    return gaussian (x1->compute (this->prec), x2->compute (this->prec)); }
};

TMPL template<typename XT,typename YT>
Certify::certify (const XT& x, const YT& y):
  rep (new complex_certify_rep<C,V> (Real_certify (x), Real_certify (y))) {}

template<typename Op,typename C,typename V>
class complexify_certify_rep: public Complex_certify_rep {
  Certify x, y;
public:
  inline complexify_certify_rep (const Certify& x2, const Certify& y2):
    x (x2), y (y2) { this->init (); }
  generic expression () const {
    return Op::op (as_generic (x), as_generic (y)); }
  Complex_type(C) compute () {
    return Op::op (x->compute (this->prec), y->compute (this->prec)); }
};

template<typename Op,typename C,typename V> inline Complex_certify
complexify_certify (const Certify& x, const Certify& y) {
  return (Complex_certify_rep*) new complexify_certify_rep<Op,C,V> (x, y); }

TMPL Complex_certify gaussian (const Certify& x, const Certify& y) {
  return complexify_certify<gaussian_op> (x, y); }
TMPL Complex_certify polar (const Certify& x, const Certify& y) {
  return complexify_certify<polar_op> (x, y); }

/******************************************************************************
* From complex numbers to real or complex numbers
******************************************************************************/

template<typename Op,typename C,typename V>
class real_unary_certify_rep: public Real_certify_rep {
  Certify z;
public:
  inline real_unary_certify_rep (const Certify& z2): z (z2) { this->init (); }
  generic expression () const { return Op::op (as_generic (z)); }
  Real_type(C) compute () { return Op::op (z->compute (this->prec)); }
};

template<typename Op,typename C,typename V> inline Real_certify
real_unary_certify (const Certify& x) {
  return (Real_certify_rep*) new real_unary_certify_rep<Op,C,V> (x); }

TMPL Real_certify Re (const Certify& z) {
  return real_unary_certify<Re_op> (z); }
TMPL Real_certify Im (const Certify& z) {
  return real_unary_certify<Im_op> (z); }
TMPL Real_certify arg (const Certify& z) {
  return real_unary_certify<arg_op> (z); }
TMPL Certify conj (const Certify& z) {
  return unary_certify<conj_op> (z); }

#undef TMPL_DEF
#undef TMPL
#undef Certify
#undef Certify_rep
#undef Center
#undef Radius
#undef Real_certify
#undef Real_certify_rep
#undef Abs_certify
#undef Abs_certify_rep
#undef Complex_certify
#undef Complex_certify_rep
}
#endif // __MML_CERTIFY_HPP
