
/******************************************************************************
* MODULE     : mml_gmp.hpp
* DESCRIPTION: Interface to gmp and mpfr
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_GMP_HPP
#define __MML_GMP_HPP
#include <gmp.h>
#include <mpfr.h>
#include "base.hpp"

namespace mml {

/******************************************************************************
* Global variables
******************************************************************************/

extern nat        mml_precision;     // current bit precision
extern mpfr_rnd_t mml_rounding_mode; // current rounding mode

inline nat set_precision (nat prec) {
  nat old= mml_precision; mml_precision= prec; return old; }
inline void reset_precision (nat prec) { mml_precision= prec; }
inline nat get_precision () { return mml_precision; }

inline mpfr_rnd_t set_rounding_mode (mpfr_rnd_t mode) {
  mpfr_rnd_t old= mml_rounding_mode; mml_rounding_mode= mode; return old; }
inline void reset_rounding_mode (mpfr_rnd_t mode) { mml_rounding_mode= mode; }
inline mpfr_rnd_t get_rounding_mode () { return mml_rounding_mode; }

/******************************************************************************
* Extra routines
******************************************************************************/

#define BITS_PER_LIMB __GMP_BITS_PER_MP_LIMB
#define DEFAULT_LOW   __GMP_BITS_PER_MP_LIMB

class string;
extern string mpfr_to_string (const mpz_t x);
extern string mpfr_to_string (const mpfr_t x, nat bits= 0);
extern string zero_to_string (const mpfr_t err);
extern ostream& operator << (ostream& out, const mpfr_t x);

inline void mpz_mul_2si (mpz_t dest, const mpz_t arg, long int shift) {
  if (shift >= 0) mpz_mul_2exp (dest, arg, shift);
  else mpz_fdiv_q_2exp (dest, arg, -shift);
}
inline void mpq_mul_2si (mpq_t dest, const mpq_t arg, long int shift) {
  if (shift >= 0) mpq_mul_2exp (dest, arg, shift);
  else mpq_div_2exp (dest, arg, -shift);
}

inline nat mpfr_prec_round_up (nat prec) {
  return BITS_PER_LIMB * ((prec + BITS_PER_LIMB - 1) / BITS_PER_LIMB); }
inline mp_exp_t mpfr_get_step (const mpfr_t arg) {
  return mpfr_get_exp (arg) - mpfr_get_prec (arg); }
inline mp_exp_t mpfr_safe_get_step (const mpfr_t arg) {
  if (mpfr_sgn (arg) == 0) return mpfr_get_emin ();
  else return mpfr_get_exp (arg) - mpfr_get_prec (arg); }
inline mp_exp_t mpfr_safe_get_exp (const mpfr_t arg) {
  if (mpfr_sgn (arg) == 0) return mpfr_get_emin ();
  else return mpfr_get_exp (arg); }

}
#endif // __MML_GMP_HPP
