
/******************************************************************************
* MODULE     : complexify.hpp
* DESCRIPTION: Complexified numbers
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_COMPLEXIFY_HPP
#define __MML_COMPLEXIFY_HPP
#include "string.hpp"
#include "mml_gmp.hpp"
#include "type_props.hpp"
namespace mml {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define Complexify complexify<C>
TMPL class complexify;
TMPL inline C Re (const Complexify& z);
TMPL inline C Im (const Complexify& z);

/******************************************************************************
* The complexify type
******************************************************************************/

TMPL_DEF
class complexify {
MML_ALLOCATORS;
private:
  C re;
  C im;
public:
  inline complexify (): re (0), im (0) {}
  template<typename T> inline complexify (const T& x):
    re (x), im (0) {}
  template<typename T> inline complexify (const complexify<T>& z):
    re (Re (z)), im (Im (z)) {}
  template<typename XT,typename YT> inline
  complexify (const XT& x, const YT& y):
    re (x), im (y) {}
  friend C Re LESSGTR (const Complexify& z);
  friend C Im LESSGTR (const Complexify& z);

  inline Complexify& operator += (const Complexify& z);
  inline Complexify& operator -= (const Complexify& z);
  inline Complexify& operator *= (const C& x);
  inline Complexify& operator /= (const C& x);
  inline Complexify& operator *= (const Complexify& z);
  inline Complexify& operator /= (const Complexify& z);
  inline Complexify& operator <<= (int shift);
  inline Complexify& operator >>= (int shift);
};

STYPE_TO_TYPE(TMPL,scalar_type,Complexify,C);
STYPE_TO_TYPE(TMPL,real_type,Complexify,C);
STYPE_TO_TYPE(TMPL,abs_type,Complexify,C);
STYPE_TO_TYPE(TMPL,center_type,Complexify,complexify<Center_type(C) >);
STYPE_TO_TYPE(TMPL,radius_type,Complexify,Radius_type(C));

TMPL
struct constant_helper<Complexify > {
  static inline Complexify nan () {
    return Complexify (Nan (C), Nan (C)); }
  static inline Complexify infinity () {
    return Complexify (Infinity (C), Infinity (C)); }
  static inline Complexify fuzz () {
    return Complexify (Fuzz (C), Fuzz(C)); }
  static inline Complexify log2 () { return Complexify (Log2 (C), 0); }
  static inline Complexify pi () { return Complexify (Pi (C), 0); }
  static inline Complexify euler () { return Complexify (Euler (C), 0); }
  static inline Complexify imaginary () { return Complexify (C (0), C (1)); }
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C Re (const Complexify& z) { return z.re; }
TMPL inline C Im (const Complexify& z) { return z.im; }

TMPL inline bool is_exact_zero (const Complexify& z) {
  return is_exact_zero (Re (z)) && is_exact_zero (Im (z)); }
TMPL inline bool is_number (const Complexify& z) {
  return is_number (Re (z)) && is_number (Im (z)); }
TMPL inline bool is_nan (const Complexify& z) {
  return is_nan (Re (z)) || is_nan (Im (z)); }
TMPL inline bool is_infinity (const Complexify& z) {
  return !is_nan (z) && (is_infinity (Re (z)) || is_infinity (Im (z))); }
TMPL inline bool is_fuzz (const Complexify& z) {
  return !is_nan (z) && (is_fuzz (Re (z)) || is_fuzz (Im (z))); }

TMPL inline Complexify copy (const Complexify& z, nat prec) {
  return Complexify (copy (Re (z), prec), copy (Im (z), prec)); }
TMPL inline nat precision (const Complexify& z) {
  return max (precision (Re (z)), precision (Im (z))); }
TMPL inline mp_exp_t expo (const Complexify& z) {
  return max (exponent (Re (z)), exponent (Im (z))); }
TMPL inline mp_exp_t step (const Complexify& z) {
  return max (step (Re (z)), step (Im (z))); }

TMPL inline nat
hash (const Complexify& z) {
  nat h= hash (Re (z));
  return (h<<1) ^ (h<<3) ^ (h<<9) ^ (h>>29) ^ hash (Im (z));
}

TMPL generic
as_generic (const Complexify& z) {
  if (is_nan (z)) return "NaN";
  if (is_fuzz (z)) return "Fuzz";
  if (is_infinity (z)) return "Infty";
  if (is_exact_zero (z)) return 0;
  if (Re (z) != 0 && Im (z) == 0) return as_generic (Re (z));
  if (Re (z) == 0 && Im (z) != 0) return times (as_generic (Im (z)), "mathi");
  return plus (as_generic (Re (z)), times (as_generic (Im (z)), "mathi"));
}

TMPL generic
as_generic (const Complexify& z, nat dd) {
  if (is_nan (z)) return "NaN";
  if (is_fuzz (z)) return "Fuzz";
  if (is_infinity (z)) return "Infty";
  if (is_exact_zero (z)) return 0;
  if (Re (z) != 0 && Im (z) == 0)
    return as_generic (Re (z), dd);
  if (Re (z) == 0 && Im (z) != 0)
    return times (as_generic (Im (z), dd), "mathi");
  return plus (as_generic (Re (z), dd),
	       times (as_generic (Im (z), dd), "mathi"));
}

PRINT_SUGAR(TMPL,Complexify)

TMPL inline bool
operator == (const Complexify& z1, const Complexify& z2) {
  return (Re (z1) == Re (z2)) && (Im (z1) == Im (z2));
}

TMPL inline bool
operator != (const Complexify& z1, const Complexify& z2) {
  return (Re (z1) != Re (z2)) || (Im (z1) != Im (z2));
}

TMPL inline int
var_sign (const Complexify& z) {
  if (z == 0) return 0; else return 1;
}

/******************************************************************************
* Polar coordinates and complex conjugation
******************************************************************************/

TMPL inline Complexify
gaussian (const C& re, const C& im= 0) {
  return Complexify (re, im);
}

TMPL inline Complexify
polar (const C& r, const C& t) {
  return Complexify (r * cos (t), r * sin (t));
}

TMPL inline C
abs (const Complexify& z) {
  return hypot (Re (z), Im (z));
}

TMPL inline C
arg (const Complexify& z) {
  return atan2 (Im (z), Re (z));
}

TMPL inline C
norm (const Complexify& z) {
  return square (Re (z)) + square (Im (z));
}

TMPL inline Complexify
conj (const Complexify& z) {
  return Complexify (Re (z), -Im (z));
}

TMPL inline Complexify
times_i (const Complexify& z) {
  return Complexify (-Im (z), Re (z));
}

TMPL inline Complexify
over_i (const Complexify& z) {
  return Complexify (Im (z), -Re (z));
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Complexify
operator + (const Complexify& z1, const Complexify& z2) {
  return Complexify (Re (z1) + Re (z2), Im (z1) + Im (z2));
}

TMPL inline Complexify
operator - (const Complexify& z) {
  return Complexify (-Re (z), -Im (z));
}

TMPL inline Complexify
operator - (const Complexify& z1, const Complexify& z2) {
  return Complexify (Re (z1) - Re (z2), Im (z1) - Im (z2));
}

TMPL inline Complexify
operator * (const C& x, const Complexify& z) {
  return Complexify (x * Re (z), x * Im (z));
}

TMPL inline Complexify
operator * (const Complexify& z, const C& x) {
  return Complexify (Re (z) * x, Im (z) * x);
}

TMPL inline Complexify
operator / (const Complexify& z, const C& x) {
  return Complexify (Re (z) / x, Im (z) / x);
}

TMPL inline Complexify
operator / (const C& x, const Complexify& z) {
  C a= x / norm (z);
  return Complexify (a * Re (z), -a * Im (z));
}

TMPL inline Complexify
operator * (const Complexify& z1, const Complexify& z2) {
  return Complexify (Re (z1) * Re (z2) - Im (z1) * Im (z2),
		     Re (z1) * Im (z2) + Im (z1) * Re (z2));
}

TMPL inline Complexify
operator / (const Complexify& z1, const Complexify& z2) {
  C a= norm (z2);
  return Complexify ((Re (z1) * Re (z2) + Im (z1) * Im (z2)) / a,
		     (Im (z1) * Re (z2) - Re (z1) * Im (z2)) / a);
}

TMPL inline Complexify
operator << (const Complexify& z, int shift) {
  return Complexify (Re (z) << shift, Im (z) << shift);
}

TMPL inline Complexify
operator >> (const Complexify& z, int shift) {
  return Complexify (Re (z) >> shift, Im (z) >> shift);
}

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Complexify&
Complexify::operator += (const Complexify& z) {
  re += z.re;
  im += z.im;
  return *this;
}

TMPL inline Complexify&
Complexify::operator -= (const Complexify& z) {
  re -= z.re;
  im -= z.im;
  return *this;
}

TMPL inline Complexify&
Complexify::operator *= (const C& x) {
  re *= x;
  im *= x;
  return *this;
}

TMPL inline Complexify&
Complexify::operator /= (const C& x) {
  re /= x;
  im /= x;
  return *this;
}

TMPL inline Complexify&
Complexify::operator *= (const Complexify& z) {
  C RE= re, zRE= z.re;
  re= re * z.re - im * z.im;
  im= RE * z.im + im * zRE;
  return *this;
}

TMPL inline Complexify&
Complexify::operator /= (const Complexify& z) {
  C RE= re;
  C a = norm (z);
  re= (re * z.re + im * z.im) / a;
  im= (im * z.re - RE * z.im) / a;
  return *this;
}

TMPL inline Complexify&
Complexify::operator <<= (int shift) {
  re <<= shift;
  im <<= shift;
  return *this;
}

TMPL inline Complexify&
Complexify::operator >>= (int shift) {
  re >>= shift;
  im >>= shift;
  return *this;
}

/******************************************************************************
* Exponentiation and logarithms
******************************************************************************/

TMPL Complexify
sqrt (const Complexify& z) {
  if (Re (z) > 0) {
    C r= sqrt ((hypot (Re (z), Im (z)) + Re (z)) / C(2));
    return Complexify (r, (Im (z) / r) / C(2));
  }
  else {
    C a= hypot (Re (z), Im (z));
    if (a == 0) return 0;
    else {
      C i = sqrt ((a - Re (z)) / C(2));
      if (Im (z) < 0) i= -i;
      return Complexify ((Im (z) / i) / C(2), i);
    }
  }
}

TMPL inline Complexify
exp (const Complexify& z) {
  return polar (exp (Re (z)), Im (z));
}

TMPL inline Complexify
log (const Complexify& z) {
  return Complexify (log (abs (z)), arg (z));
}

TMPL inline Complexify
pow (const Complexify& z, const Complexify& y) {
  C r= log (abs (z));
  C t= arg (z);
  return polar (exp (r * Re (y) - Im (y) * t),
		Im (y) * r + Re (y) * t);
}

TMPL inline Complexify
pow (const Complexify& z, const C& y) {
  return exp (y * log (z));
}

TMPL inline Complexify
pow (const C& z, const Complexify& y) {
  return exp (y * log (z));
}

TMPL inline Complexify
cos (const Complexify& z) {
  return Complexify (cos (Re (z)) * cosh (Im (z)),
		     -sin (Re (z)) * sinh (Im (z)));
}

TMPL inline Complexify
sin (const Complexify& z) {
  return Complexify (sin (Re (z)) * cosh (Im (z)),
		     cos (Re (z)) * sinh (Im (z)));
}

TMPL inline Complexify
acos (const C& z) {
  return over_i (acosh (z));
}

TMPL inline Complexify
asin (const C& z) {
  return over_i (asinh (times_i (z)));
}

TMPL inline Complexify
atan (const C& z) {
  return over_i (atanh (times_i (z)));
}

TMPL inline Complexify
cosh (const Complexify& z) {
  return Complexify (cosh (Re (z)) * cos (Im (z)),
		     sinh (Re (z)) * sin (Im (z)));
}

TMPL inline Complexify
sinh (const Complexify& z) {
  return Complexify (sinh (Re (z)) * cos (Im (z)),
		     cosh (Re (z)) * sin (Im (z)));
}

/******************************************************************************
* Miscellaneous
******************************************************************************/

TMPL inline C
rnd_error (const Complexify& z) {
  return add_U (rnd_error (Re (z)), rnd_error (Im (z)));
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Complexify,C);
EQUAL_INT_SUGAR(TMPL,Complexify,Complexify);

#undef TMPL_DEF
#undef TMPL
#undef Complexify
} // namespace mml
#endif // __MML_COMPLEXIFY_HPP
