
/******************************************************************************
* MODULE     : dense_polynomial.hpp
* DESCRIPTION: Dense univariate polynomials
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DENSE_POLYNOMIAL_HPP
#define __MML_DENSE_POLYNOMIAL_HPP
#include "polynomial.hpp"
#include "jet.hpp"
namespace mml {
#define TMPL template<typename C, typename V>
#define Variant dense_polynomial<V>
#define Polynomial polynomial<C,Variant >
#define Polynomial_rep polynomial_rep<C,Variant >
TMPL inline int deg (const Polynomial& P);
TMPL inline nat dim (const Polynomial& P);
TMPL inline const C* seg (const Polynomial& P);

/******************************************************************************
* Dense polynomial variants
******************************************************************************/

template<typename V> class dense_polynomial {};

/******************************************************************************
* Dense polynomials
******************************************************************************/

template<typename C, typename V>
class Polynomial_rep REP_STRUCT {
  C*  a; // size of a can be larger than n due to normalization
  nat n;

public:
  inline void normalize () { while ((n != 0) && (a[n-1] == C(0))) n--; }
  inline polynomial_rep (C* a2, nat n2): a (a2), n (n2) { normalize (); }
  inline ~polynomial_rep () { delete[] a; }
  void extend (nat l) {
    C* b= new C[l];
    jet_copy (b, a, n);
    jet_clear (b+n, l-n);
    delete[] a;
    a= b;
    n= l;
  }

  friend class Polynomial;
  friend int deg LESSGTR (const Polynomial& P);
  friend nat dim LESSGTR (const Polynomial& P);
  friend const C* seg LESSGTR (const Polynomial& P);
};

template<typename C, typename V>
class Polynomial {
INDIRECT_PROTO_2 (polynomial, polynomial_rep, C, dense_polynomial<V>)
public:
  inline polynomial () {
    rep= new Polynomial_rep (new C[0], 0);
  }
  template<typename T> inline polynomial (const T& x) {
    C* a= new C[1];
    a[0]= x;
    rep= new Polynomial_rep (a, 1);
  }
  template<typename T> inline
  polynomial (const polynomial<T,dense_polynomial<V> >& P) {
    C* a= new C[dim (P)];
    jet_convert (a, seg (P), dim (P));
    rep= new Polynomial_rep (a, dim (P));
    rep->normalize ();
  }
  template<typename T> polynomial (const T& x, nat n) {
    C* a= new C[n+1];
    jet_clear (a, n);
    a[n]= x;
    rep= new Polynomial_rep (a, n+1);
  }
  inline polynomial (C* a, nat n) {
    rep= new Polynomial_rep (a, n);
  }
  polynomial (const iterator<C>& it) {
    nat i, l=0;
    rep= new Polynomial_rep (new C[0], 0);
    for (i=0; busy (it); i++, ++it) {
      if (i >= l) {
	l= max ((nat) 1, l<<1);
	rep->extend (l);
      }
      rep->a[i]= *it;
    }
    rep->normalize ();
  }
  inline C operator[] (nat i) const { return rep->a[i]; }
  Polynomial& operator += (const Polynomial& P);
  Polynomial& operator -= (const Polynomial& P);
  Polynomial& operator *= (const C& c);
  Polynomial& operator /= (const C& c);
};
INDIRECT_IMPL_2 (polynomial, polynomial_rep,
		 typename C, C, typename V, dense_polynomial<V>)

TMPL inline nat dim (const Polynomial& P) { return P->n; }
TMPL inline int deg (const Polynomial& P) { return ((int) P->n) - 1; }
TMPL inline const C* seg (const Polynomial& P) { return P->a; }

/******************************************************************************
* In place arithmetic
******************************************************************************/

TMPL Polynomial&
Polynomial::operator += (const Polynomial& P) {
  nat n1= rep->n, n2= dim (P);
  if (n1 < n2) rep -> extend (n2);
  jet_add<C> (rep->a, seg (P), n2);
  if (n1 == n2) rep -> normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator -= (const Polynomial& P) {
  nat n1= rep->n, n2= dim (P);
  if (n1 < n2) rep->extend (n2);
  jet_sub<C> (rep->a, seg (P), n2);
  if (n1 == n2) rep->normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator *= (const C& c) {
  jet_mul<C> (rep->a, c, rep->n);
  rep->normalize ();
  return *this;
}

TMPL Polynomial&
Polynomial::operator /= (const C& c) {
  jet_div<C> (rep->a, c, rep->n);
  rep->normalize ();
  return *this;
}

/******************************************************************************
* Basic arithmetic operations
******************************************************************************/

TMPL Polynomial
operator + (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (dim (P1), dim (P2));
  nat n= max (dim (P1), dim (P2));
  C* r= new C[n];
  jet_add<C> (r, seg (P1), seg (P2), m);
  if (dim (P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (dim (P2) > m) jet_copy<C> (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n);
}

TMPL Polynomial
operator - (const Polynomial& P1, const Polynomial& P2) {
  nat m= min (dim (P1), dim (P2));
  nat n= max (dim (P1), dim (P2));
  C* r= new C[n];
  jet_sub<C> (r, seg (P1), seg (P2), m);
  if (dim (P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (dim (P2) > m) jet_neg<C> (r+m, seg(P2)+m, n-m);
  return Polynomial (r, n);
}

TMPL Polynomial
operator - (const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_neg<C> (r, seg (P), n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator * (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 == 0 || n2 == 0) return Polynomial ();
  C* r= new C[n1+n2-1];
  jet_mul<C,V> (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n1+n2-1);
}

TMPL Polynomial
operator * (const Polynomial& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_mul<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator * (const C& c, const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_mul<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator / (const Polynomial& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_div<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
quo (const Polynomial& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_quo<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
rem (const Polynomial& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_rem<C> (r, seg (P), c, n);
  return Polynomial (r, n);
}

TMPL Polynomial
operator / (const Polynomial& P1, const Polynomial& P2) PARTIAL {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  C* q= new C[n1-n2+1];
  C* r= new C[n1];
  jet_copy (r, seg (P1), n1);
  jet_quo_rem<C,V> (q, r, seg (P2), n1, n2);
  domain_check ((Polynomial (r, n1) == 0)); // division must be exact
  return Polynomial (q, n1-n2+1);
}

TMPL Polynomial
quo (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2 || n2 == 0) return Polynomial ();
  C* q= new C[n1-n2+1];
  C* r= new C[n1];
  jet_copy (r, seg (P1), n1);
  jet_quo_rem<C,V> (q, r, seg (P2), n1, n2);
  return Polynomial (q, n1-n2+1);
}

TMPL Polynomial
rem (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2 || n2 == 0) return P1;
  C* q= new C[n1-n2+1];
  C* r= new C[n1];
  jet_copy (r, seg (P1), n1);
  jet_quo_rem<C,V> (q, r, seg (P2), n1, n2);
  return Polynomial (r, n1);
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL int
sign (const Polynomial& P) {
  nat n= dim (P);
  if (n == 0) return 0;
  return sign (P[n-1]);
}

TMPL int
compare (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 > n2) return sign (P1[n1-1]);
  if (n2 > n1) return -sign (P2[n2-1]);
  return jet_cmp (seg (P1), seg (P2), n1);
}

TMPL int
unequal (const Polynomial& P1, const Polynomial& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 != n2) return 1;
  for (nat i=0; i<n1; i++)
    if (P1[i] != P2[i]) return 1;
  return 0;
}

/******************************************************************************
* Differentiation, integration and composition
******************************************************************************/

TMPL Polynomial
derive (const Polynomial& P) {
  nat n= dim (P);
  if (n <= 1) return 0;
  C* r= new C[n-1];
  jet_derive (r, seg (P), n);
  return Polynomial (r, n-1);
}

TMPL Polynomial
xderive (const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_xderive (r, seg (P), n);
  return Polynomial (r, n);
}

TMPL Polynomial
integrate (const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n+1];
  jet_integrate (r, seg (P), n);
  return Polynomial (r, n+1);
}

template<typename C,typename V,typename K> Polynomial
compose (const Polynomial& P1, const polynomial<K,Variant >& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 <= 1) return P1;
  if (n2 == 0) return P1[0];
  nat n= (n1-1) * (n2-1) + 1;
  C* r= new C[n];
  jet_dicho_compose<C,V,K> (r, seg (P1), seg (P2), n1, n2);
  return Polynomial (r, n);
}

TMPL Polynomial
shift (const Polynomial& P, const C& sh) {
  nat n= dim (P);
  if (n <= 1) return P;
  C q[2]; q[0]= sh; q[1]= 1;
  C* r= new C[n];
  jet_dicho_compose<C,V> (r, seg (P), q, n, 2);
  return Polynomial (r, n);
}

TMPL Polynomial
shift (const Polynomial& P, int i) {
  return shift (P, C (i));
}

template<typename C,typename V,typename K> Polynomial
q_difference (const Polynomial& P, const K& q) {
  nat n= dim (P);
  if (n <= 1) return P;
  C* r= new C[n];
  jet_q_difference (r, seg (P), q, n);
  return Polynomial (r, n);
}

TMPL Polynomial
dilate (const Polynomial& P, nat p) {
  nat n= dim (P);
  if (n <= 1) return P;
  nat k= (n-1)*p + 1;
  C* r= new C[k];
  jet_dilate (r, seg (P), p, n);
  return Polynomial (r, k);
}

/******************************************************************************
* Miscellaneous
******************************************************************************/

TMPL Polynomial
range (const Polynomial& P, nat start, nat end) {
  C* r= new C[end-start];
  for (nat i=start; i<end; i++) r[i-start]= P[i];
  return Polynomial (r, end-start);
}

TMPL Polynomial
reverse (const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  for (nat i=0; i<n; i++) r[i]= P[n-1-i];
  return Polynomial (r, n);
}

TMPL Polynomial
square_roots (const Polynomial& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_square_roots<C,V> (r, seg (P), n);
  return Polynomial (r, n);
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Polynomial,C);
SHIFT_INT_SUGAR(TMPL,Polynomial,C);
COMPARE_SUGAR(TMPL,Polynomial);
COMPARE_INT_SUGAR(TMPL,Polynomial,Polynomial);

#undef TMPL
#undef Variant
#undef Polynomial
#undef Polynomial_rep
} // namespace mml
#endif // __MML_DENSE_POLYNOMIAL_HPP
