
/******************************************************************************
* MODULE     : block.hpp
* DESCRIPTION: header for low level matricial computations
* COPYRIGHT  : (C) 2004  Olivier Ruatta
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__BLOCK__HPP
#define __MML__BLOCK__HPP
#include "segment.hpp"

namespace mml {
#define TMPL template<class C>

/******************************************************************************
* Operations on blocks
******************************************************************************/

TMPL void
block_set (C* dest, const C& c, nat rows, nat cols) {
  segment_clear (dest, rows*cols);
  nat m= min (rows, cols);
  while (m != 0) {
    *dest= c;
    dest += cols+1; m--;
  }
}

TMPL void
block_extraction (C* dest, const C* s,
		  nat rows, nat r_start, nat r_end,
		  nat cols, nat c_start, nat c_end)
{
  s += c_start + r_start*cols; r_end -= r_start; c_end -= c_start;
  for (nat row= 0; row<r_end; row++) {
    segment_copy (dest, s, c_end);
    dest += c_end;
    s += cols;
  }
}

TMPL void
block_duplicate (C* dest, nat dr, nat dc, nat ddr, nat ddc,
		 C* b, nat br, nat bc, nat dbr, nat fbr, nat dbc, nat fbc)
{
  nat sr = fbr - dbr + 1;
  nat sc = fbc - dbc + 1;
  assert ((dbr < fbc) && (fbc < bc));
  assert ((ddr + sr - 1 < dr) && (dbc + sc - 1 < dc));
  C* id = dest + ddr * dc;
  C* ib = b + dbr * bc;
  for (nat i = 0; i < sr; i++) {
    id += ddc;
    ib += dbc;
    for (nat j = 0; j < sc; j++) {
      *id = *ib;
      id++; ib++;
    }
    id += dc - (ddc + sc);
    ib += bc - fbc - 1;
  }
}

TMPL void
block_transpose (C* dest, const C* s, nat rows, nat cols) {
  nat tcols=0;
  while (tcols != cols) {
    nat trows=0;
    while (trows != rows) {
      *dest = *(s + trows*cols);
      dest++; trows++;
    }
    s++; tcols++;
  }
}

TMPL void
block_naive_mul (C* dest, const C* s1, const C* s2, nat r1, nat c1, nat c2) {
  C tmp;
  nat i, j, k;
  for (i=0; i<r1; i++)
    for (j=0; j<c2; j++) {
      tmp= 0;
      for (k=0; k<c1; k++)
	tmp += s1[i*c1+k] * s2[j+k*c2];
      dest[i*c2+j]= tmp;
    }
}

/******************************************************************************
* Operations on rows and columns
******************************************************************************/

TMPL void
row_in_place_swap (C* s, nat i, nat j, nat r, nat c) {
  assert (i<r && j<r);
  C* ri= s + i*c;
  C* rj= s + j*c;
  C aux;
  while (c > 0) { 
    aux= *ri; *ri= *rj; *rj= aux;
    ri++; rj++; c--;
  }
}

TMPL void
col_in_place_swap (C* s, nat i, nat j, nat r, nat c) {
  assert (i<c && j<c);
  C* ci= s+i;
  C* cj= s+j;
  C aux;
  while (r>0) {
    aux= *ci; *ci= *cj; *cj= aux;
    ci+=c; cj+=c; r--;
  }
}

template<typename Op,typename C,typename X> void
row_in_place_extrary (C* s, const X& x, nat i, nat r, nat c) {
  assert (i<r);
  C* it= s + i*c;
  while (c>0) {
    Op::in_place (*it, x);
    it++; c--;
  }
}

template<typename Op,typename C,typename X> void
col_in_place_extrary (C* s, const X& x, nat i, nat r, nat c) {
  assert (i<c);
  C* it= s+i;
  while (r>0) {
    Op::in_place (*it, x);
    it+=c; r--;
  }
}

TMPL void row_in_place_mul (C* s, const C& sc, nat i, nat r, nat c) {
  row_in_place_extrary<sc_mul_op<false> > (s, sc, i, r, c); }
TMPL void row_in_place_div (C* s, const C& sc, nat i, nat r, nat c) {
  row_in_place_extrary<sc_div_op<false> > (s, sc, i, r, c); }
TMPL void col_in_place_mul (C* s, const C& sc, nat i, nat r, nat c) {
  col_in_place_extrary<sc_mul_op<false> > (s, sc, i, r, c); }
TMPL void col_in_place_div (C* s, const C& sc, nat i, nat r, nat c) {
  col_in_place_extrary<sc_div_op<false> > (s, sc, i, r, c); }

TMPL void
row_combine_add (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) {
  assert (i<r && j<r);
  C* iti= s + i*c;
  C* itj= s + j*c;
  while (c>0) {
    *iti= si * (*iti) + sj * (*itj);
    iti++; itj++; c--;
  }
}

TMPL inline void
row_combine_sub (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) {
  row_combine_add (s, i, si, j, -sj, r, c);
}

TMPL void
col_combine_add (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) {
  assert (i<c && j<c);
  C* iti= s+i, itj= s+j;
  while (r>0) {
    *iti = si * (*iti) + sj * (*itj);
    iti+=c; itj+=c; r--;
  }
}

TMPL inline void
col_combine_sub (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) {
  col_combine_add (s, i, si, j, -sj, r, c);
}

#undef TMPL
}//namespace mml
#endif //__MML__BLOCK__HPP
