
/******************************************************************************
* MODULE     : fft_root.hpp
* DESCRIPTION: arithmetic on 2^k-th roots of unity used by the FFT
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__FFT_ROOT__HPP
#define __MML__FFT_ROOT__HPP
#include "modular.hpp"

namespace mml {
#define TMPL template<class C>

/******************************************************************************
* Primitive roots of unity
******************************************************************************/

TMPL
class fft_root {
MML_ALLOCATORS;
public:
  C u;
  inline fft_root (): u (1) {}
  fft_root (nat n);
  inline fft_root (const C& c): u (c) {}
  inline fft_root<C>& operator = (const fft_root<C>& x) {
    u= x.u; return *this; }
  inline bool operator == (const fft_root<C>& v) { return u == v.u; }
  inline bool operator != (const fft_root<C>& v) { return u != v.u; }
};

TMPL C inline
operator * (const C& c, const fft_root<C>& u) {
  return c * u.u;
}

TMPL fft_root<C> inline
operator * (const fft_root<C>& u, const fft_root<C>& v) {
  return fft_root<C> (u.u * v.u);
}

TMPL inline generic as_generic (const fft_root<C>& u) {
  return as_generic (u.u); }
TMPL inline ostream& operator << (ostream& out, const fft_root<C>& u) {
  return out << u.u; }

/******************************************************************************
* Special operations for FFT
******************************************************************************/

static nat cross_count;

TMPL inline void
fft_cross (C* c1, C* c2, const fft_root<C>* u) {
  C temp= (*c2) * (*u);
  *c2   = (*c1) - temp;
  *c1   = (*c1) + temp;
  cross_count++;
}

TMPL inline void
inverse_fft_cross (C* c1, C* c2, const fft_root<C>* u) {
  C temp= *c2;
  *c2   = ((*c1) - temp) * (*u);
  *c1   = (*c1) + temp;
}

TMPL inline void
fft_shift (C* dest, nat n) {
  nat i= n;
  C u= invert (C((int) n));
  while (i!=0) {
    *dest= u * (*dest);
    dest++; i--;
  }
}

/******************************************************************************
* A particularly nice modulus for FFT: p= 3*2^30+1
******************************************************************************/

#define FFTP nat_value<0x0c0000001>

template<>
fft_root<modular<FFTP> >::fft_root (nat n) {
  modular<FFTP> v= 125;
  assert ((1<<30)%n == 0); // |n| must be a power of two
  u= binpow (v, (1<<30) / n);
}

#undef TMPL
} // namespace mml
#endif //__MML__FFT_ROOT__HPP
