
/******************************************************************************
* MODULE     : relaxed_series.hpp
* DESCRIPTION: Relaxed univariate power series
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_RELAXED_SERIES_HPP
#define __MML_RELAXED_SERIES_HPP
#include "series.hpp"
#include "dense_vector.hpp"
#include "karatsuba.hpp"
#include "fft.hpp"
namespace mml {
#define TMPL template<typename C, typename V>
#define Variant relaxed<V>
#define Series series<C,Variant >
#define Series_rep series_rep<C,Variant >
#define Vector vector<C>
#define Series_vector series<Vector,Variant >
#define Series_vector_rep series_rep<Vector,Variant >

/******************************************************************************
* Relaxed series variants
******************************************************************************/

template<typename V>
class relaxed {
public:
};

/******************************************************************************
* Setting up the different levels for relaxed multiplication
******************************************************************************/

#define TRANSFORM_NAIVE      0
#define TRANSFORM_KARATSUBA  1
#define TRANSFORM_FFT        2

template <typename C>
struct level_info {
  typedef C* C_ptr;

  nat n;          // size of blocks
  nat k;          // multiplier
  nat tsz;        // size of transformed blocks
  nat type;       // type of the transformation
  C_ptr head[2];  // buffer for transformed heads
  C_ptr tail[2];  // buffer for transformed tails

  level_info () {
    head[0]= head[1]= NULL;
    tail[0]= tail[1]= NULL;
  }
  ~level_info () {
    if (head[0] != NULL) delete[] head[0];
    if (head[1] != NULL) delete[] head[1];
    if (tail[0] != NULL) delete[] tail[0];
    if (tail[1] != NULL) delete[] tail[1];
  }
};

static nat
get_inside_multiplier (nat n) {
  if (n <= (1 << 7)) return 16;
  if (n <= (1 << 9)) return 4;
  if (n <= (1 << 10)) return 8;
  if (n <= (1 << 15)) return 16;
  if (n <= (1 << 23)) return 32;
  return 64;
}

static nat
get_border_multiplier (nat n) {
  if (n <= (1 << 4)) return 2;
  if (n <= (1 << 8)) return 4;
  if (n <= (1 << 13)) return 8;
  if (n <= (1 << 21)) return 16;
  return 32;
}

  /* Better for semi-relaxed multiplication
static nat
get_border_multiplier (nat n) {
  if (n <= (1 << 4)) return 2;
  if (n <= (1 << 6)) return 4;
  if (n <= (1 << 9)) return 8;
  if (n <= (1 << 17)) return 16;
  return 32;
}
  */

  /* Better for (semi-)relaxed mult. over F_p and k-bit compl. floats
static nat
get_border_multiplier (nat n) {
  if (n <= (1 << 3)) return 2;
  if (n <= (1 << 4)) return 4;
  if (n <= (1 << 8)) return 8;
  if (n <= (1 << 16)) return 16;
  return 32;
}
  */

static vector<nat>
determine_sizes (nat n) {
  if (n <= 16) return vector<nat> (1, 1);
  vector<nat> v;
  nat k= get_inside_multiplier (n);
  // nat k= get_inside_multiplier (n*2); for F_p and k-bit compl. floats
  n /= k;
  while (n != 1) {
    k= get_border_multiplier (n);
    v << k;
    n /= k;
  }
  v << 1;
  for (nat i=0; i<dim(v)/2; i++)
    swap (v (i), v (dim(v) - 1 - i));
  return v;
}

/******************************************************************************
* Relaxed multiplication
******************************************************************************/

//static nat mul_count= 0;

TMPL
class nrelax_mul_series_rep: public Series_rep {
protected:
  Series f[2];           // the series being multiplied
  nat sh[2];             // sh[i]=1 if f[1-i] is relaxed and 0 otherwise
  nat capacity[2];       // capacity for inner transforms
  nat nr_levels;         // number of levels
  level_info<C>* xinfo;  // storage of info
  level_info<C>* info;   // information about each level

public:
  nrelax_mul_series_rep (const Series& f2, const Series& g2, nat n) {
    f[0]= f2; f[1]= g2;
    sh[0]= 1; sh[1]= 1;
    //sh[0]= 1; sh[1]= 0;
    vector<nat> v= determine_sizes (n);
    nr_levels= dim (v);
    info= new level_info<C> [nr_levels];
    nat sz= 1;
    for (nat level=0; level< nr_levels; level++) {
      bool last= (level == nr_levels - 1);
      nat n= sz * v[level];
      nat k= (last? 1: v[level+1]);
      info[level].n = n;
      info[level].k = k;
      nat tsz, type;
      if (v[level] == 1) {
	tsz = 1;
	type= TRANSFORM_NAIVE;
      }
      else if (v[level] == 2) {
	tsz = 3 * info[level-1].tsz;
	type= TRANSFORM_KARATSUBA;
      }
      else {
	tsz= 2 * n;
	type= TRANSFORM_FFT;
      }
      info[level].tsz = tsz;
      info[level].type= type;
      //cout << "Level " << level << "\n";
      //cout << "  " << "n   = " << n << "\n";
      //cout << "  " << "k   = " << k << "\n";
      //cout << "  " << "tsz = " << tsz << "\n";
      //cout << "  " << "type= " << type << "\n";
      if (last) {
	info[level].tail[0]= new C[2 * tsz];
	info[level].tail[1]= new C[2 * tsz];
	capacity[0]= 2;
	capacity[1]= 2;
      }
      else {
	if (sh[0] != 0) {
	  info[level].head[0]= new C[k * tsz];
	  info[level].tail[1]= new C[k * tsz];
	}
	if (sh[1] != 0) {
	  info[level].head[1]= new C[k * tsz];
	  info[level].tail[0]= new C[k * tsz];
	}
      }
      sz *= v[level];
      //cout << "  " << "head= " << info[level].head[0] << ", " << info[level].head[1] << "\n";
      //cout << "  " << "tail= " << info[level].tail[0] << ", " << info[level].tail[1] << "\n";
    }
    if (sh[0] == 0 && sh[1] == 0) {
      xinfo= info;
      info += nr_levels-1;
      nr_levels= 1;
    }
    else xinfo= info;
  }

  ~nrelax_mul_series_rep () { delete[] xinfo; }

  generic As_generic (const generic& z) const {
    return times (as_generic (f[0], z), as_generic (f[1], z)); }

  nat allocated (nat l) {
    if (l == 0) return 0;
    else return l + 2 * info[nr_levels-1].n; }
  void Set_precision (nat l2) { // not very nice, but OK...
    if (l2 <= this->l) return;
    nat old_allocated= allocated (this->l);
    nat new_allocated= allocated (l2);
    C* b= new C[new_allocated];
    jet_copy (b, this->a, old_allocated);
    jet_clear (b + old_allocated, new_allocated - old_allocated);
    delete[] this->a;
    this->a= b;
    this->l= l2;
  }

  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f[0], l);
    increase_precision (f[1], l); }

  void direct_transform (C* dest, nat ld, const C* src, nat ls, nat type) {
    //cout << "    ld= " << ld << ", ls= " << ls << ", type= " << type << "\n";
    if (type == TRANSFORM_NAIVE) dest[0]= src[0];
    else if (type == TRANSFORM_KARATSUBA) {
      jet_copy (dest, src, ls);
      direct_fkt (dest, ls, ld);
    }
    else {
      jet_copy (dest, src, ls);
      jet_clear (dest + ls, ls);
      fft (dest, ld, false);
    }
  }

  void inverse_transform (C* dest, nat n, nat tsz, nat type) {
    if (type == TRANSFORM_NAIVE);
    else if (type == TRANSFORM_KARATSUBA) inverse_fkt (dest, tsz, n);
    else fft (dest, tsz, true);
  }

  void direct_transform (nat which) {
    for (nat level= 0; level < nr_levels; level++) {
      typedef C* C_ptr;
      nat    n   = info[level].n;
      nat    tsh = sh[0] + sh[1];
      nat    cur = this->n + tsh - n * sh[1-which];
      if (cur % n != 0) break;
      if (sh[1-which] != 0 && this->n + 1 < n) break;
      //cout << "  Transform " << cur/n << " (" << which << ") at level " << level << "\n";
      bool   last= (level == nr_levels - 1);
      nat    k   = info[level].k;
      nat    tsz = info[level].tsz;
      nat    type= info[level].type;
      C_ptr& head= info[level].head[which];
      C_ptr& tail= info[level].tail[which];
      nat Cur= cur / n;
      nat Mod= Cur % k;
      //cout << "    Mod= " << Mod << "\n";
      if (f[which]->n < this->n + 1 && which == 1) {
	cout << ">>> n= " << this->n << " and only " << f[which]->n << " terms\n";
	assert (f[which]->n >= this->n + 1);
      }
      const C* seg= (sh[1-which] == 0?
		     f[which] (this->n, this->n + n):
		     f[which] (this->n + 1 - n, this->n + 1));
      //cout << "    seg= "; segment_print (cout, seg, n); cout << "\n";
      if (last) {
	nat Cap= capacity[which];
	//cout << "    Cur= " << cur << ", Cap= " << Cap << "\n";
	if (Cur >= Cap) {
	  nat old_cap= Cap * tsz;
	  nat new_cap= old_cap << 1;
	  C* a= new C [new_cap];
	  jet_copy (a, tail, old_cap);
	  jet_clear (a + old_cap, new_cap - old_cap);
	  delete[] tail;
	  tail= a;
	  capacity[which]= Cap << 1;
	  //cout << "    New capacity= " << capacity[which] << "\n";
	}
	//cout << "    -> tail " << Cur << "\n";
	direct_transform (tail + Cur * tsz, tsz, seg, n, type);
	//cout << "    tr= "; segment_print (cout, tail + Cur * tsz, tsz); cout << "\n";
      }
      else if (Cur < k * sh[which]) {
	//cout << "    -> head " << Mod << "\n";
	direct_transform (head + Mod * tsz, tsz, seg, n, type);
	//cout << "    tr= "; segment_print (cout, head + Mod * tsz, tsz); cout << "\n";
      }
      else if (tail != NULL) {
	//cout << "    -> tail " << Mod << "\n";
	direct_transform (tail + Mod * tsz, tsz, seg, n, type);
	//cout << "    tr= "; segment_print (cout, tail + Mod * tsz, tsz); cout << "\n";
      }
    }
  }

  inline void accumulate (C* dest, const C* s1, const C* s2, nat len) {
    //mul_count += len;
    //cout << "      s1= "; segment_print (cout, s1, len); cout << "\n";
    //cout << "      s2= "; segment_print (cout, s2, len); cout << "\n";
    segment_mul_add (dest, s1, s2, len);
    //cout << "      d = "; segment_print (cout, dest, len); cout << "\n";
  }

  C next () {
    //cout << "[" << this->n << "]"; cout.flush ();
    //cout << "Coefficient: " << this->n << "\n";
    direct_transform (0);
    direct_transform (1);
    for (nat level= 0; level < nr_levels; level++) {
      nat   tsh = sh[0] + sh[1];
      nat   cur = this->n + tsh;
      nat   n   = info[level].n;
      if (cur % n != 0) break;
      nat   Cur = cur/n;
      if (Cur < tsh) break;
      //cout << "  Product at level " << level << "\n";
      bool  last= (level == nr_levels - 1);
      nat   msh = min (sh[0], sh[1]);
      nat   k   = info[level].k;
      nat   tsz = info[level].tsz;
      nat   type= info[level].type;
      const C* h0 = info[level].head[0];
      const C* h1 = info[level].head[1];
      const C* t0 = info[level].tail[0];
      const C* t1 = info[level].tail[1];
      nat   Mod = Cur % k;

      //cout << "    tsz= " << tsz << ", l= " << 2*n-1 << "\n";
      C* acc= new C[tsz];
      segment_clear (acc, tsz);
      if (last)
	for (nat i=sh[0]; i<=Cur-sh[1]; i++)
	  accumulate (acc, t0 + i*tsz, t1 + (Cur-i)*tsz, tsz);
      else {
	if (Cur < 2 * msh * k) {
	  nat start= 1      ; if (Cur > k) start= Cur - k + 1;
	  nat end  = Cur - 1; if (Cur > k) end  = k - 1;
	  for (nat i=start; i<=end; i++)
	    accumulate (acc, h0 + i*tsz, h1 + (Cur-i)*tsz, tsz);
	}
	if (Cur >= msh * k && Mod != 0) {
	  //cout << "    Normal\n";
	  if (sh[0] != 0)
	    for (nat i=1; i<=Mod; i++)
	      accumulate (acc, h0 + i*tsz, t1 + (Mod-i)*tsz, tsz);
	  if (sh[1] != 0)
	    for (nat i=0; i<=Mod-1; i++)
	      accumulate (acc, t0 + i*tsz, h1 + (Mod-i)*tsz, tsz);
	}
      }
      inverse_transform (acc, n, tsz, type);
      //cout << "    acc= "; segment_print (cout, acc, 2*n-1); cout << "\n";
      jet_add (this->a + this->n, acc, 2*n - 1);

      if (Cur >= msh * k && Mod == 0 && !last) {
	//cout << "    Extra " << k << "\n";
	for (nat Mod= 0; Mod<k-1; Mod++) {
	  segment_clear (acc, tsz);
	  for (nat i=Mod+1; i<k; i++) {
	    if (sh[0] != 0)
	      accumulate (acc, h0 + i*tsz, t1 + (k+Mod-i)*tsz, tsz);
	    if (sh[1] != 0)
	      accumulate (acc, t0 + i*tsz, h1 + (k+Mod-i)*tsz, tsz);
	  }
	  inverse_transform (acc, n, tsz, type);
	  jet_add (this->a + this->n + Mod*n, acc, 2*n - 1);
	}
      }
      delete[] acc;
    }
    return this->a [this->n];
  }
};

TMPL Series
nrelax_mul (const Series& f, const Series& g, nat n) {
  return (Series_rep*) new nrelax_mul_series_rep<C,V> (f, g, n);
}

/******************************************************************************
* Top level interface for relaxed multiplication
******************************************************************************/

TMPL
class relaxed_mul_series_rep: public binary_series_rep<mul_op,C,Variant > {
protected:
  Series prod;
  nat    N;
public:
  inline relaxed_mul_series_rep (const Series& f, const Series& g):
    binary_series_rep<mul_op,C,Variant > (f, g),
    prod (nrelax_mul (f, g, 1)), N (1) {}
  C next () { return prod [this->n]; }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (prod, l);
    if (l < 2*N) return;
    while (l >= 2*N) N= 2*N;
    //while (l >= 2*N) N= 4*N;
    prod= nrelax_mul (this->f, this->g, N);
  }
};

TMPL Series
operator * (const Series& f, const Series& g) {
  //return nrelax_mul (f, g, 64);
  return (Series_rep*) new relaxed_mul_series_rep<C,V> (f, g);
}

#undef TMPL
#undef Variant
#undef Series
#undef Series_rep
#undef Vector
#undef Series_vector
#undef Series_vector_rep
} // namespace mml
#endif // __MML_SERIES_HPP
